# mypy: disable-error-code="override"

"""
Get metadata for the courts covered by the service
"""

import pathlib
from datetime import date
from enum import Enum
from functools import cached_property
from re import compile
from typing import Any, Dict, Optional

from markdown_it import MarkdownIt
from mdit_py_plugins.attrs import attrs_plugin
from ruamel.yaml import YAML

from ds_caselaw_utils.types.courts_schema_autogen import (
    RawCourt,
    RawCourtRepository,
    RawJurisdiction,
)

from .types import CourtCode, CourtParam, JurisdictionCode, NeutralCitationPattern

md = MarkdownIt("commonmark", {"breaks": True, "html": True}).use(attrs_plugin)


class FormatMapDict(dict[str, Any]):
    def __missing__(self, key: str) -> str:
        return "{" + key + "}"


class InstitutionType(Enum):
    COURT = "court"
    TRIBUNAL = "tribunal"


class Jurisdiction:
    def __init__(self, data: RawJurisdiction):
        self.code: JurisdictionCode = JurisdictionCode(data["code"])
        self.name: str = data["name"]
        self.prefix: Optional[str] = data.get("prefix")


class Court:
    def __init__(self, data: RawCourt, type: InstitutionType) -> None:
        self.canonical_param: Optional[CourtParam]
        self.param_aliases: list[CourtParam]
        self.code: CourtCode = CourtCode(data["code"])
        self.name: str = data["name"]
        self.long_name: str = data.get("long_name") or data["name"]
        self.grouped_name: str = data.get("grouped_name") or data["name"]
        self.link: str = data["link"]
        self.identifier_iri: str = data.get("identifier_iri") or data["link"]
        self.ncn_pattern: Optional[NeutralCitationPattern] = (
            NeutralCitationPattern(compile(data["ncn_pattern"])) if "ncn_pattern" in data else None
        )
        self.ncn_examples: Optional[list[str]] = data["ncn_examples"] if "ncn_examples" in data else None
        if "param" in data:
            self.canonical_param = CourtParam(data["param"])
            self.param_aliases = [CourtParam(data["param"])] + [
                CourtParam(extra_param) for extra_param in data.get("extra_params", [])
            ]
        else:
            self.canonical_param = None
            self.param_aliases = []

        self.type = type

        self.start_year: Optional[int] = data.get("start_year")
        self.end_year: int = data.get("end_year") or date.today().year
        self.jurisdictions: list[Jurisdiction] = [
            Jurisdiction(jurisdiction_data) for jurisdiction_data in data.get("jurisdictions", [])
        ]

    def get_jurisdiction(self, code: str) -> Optional[Jurisdiction]:
        return next((j for j in self.jurisdictions if j.code == code), None)

    def expand_jurisdictions(self) -> list["Court"]:
        return [self] + [CourtWithJurisdiction(self, jurisdiction) for jurisdiction in self.jurisdictions]

    def render_markdown_text(self, type: str, context: Dict[str, Any] = {}) -> Optional[str]:
        if not self.canonical_param:
            return None

        filename = self.canonical_param.replace("/", "_")
        description_md_file_path = pathlib.Path(__file__).parent / f"data/markdown/{type}/{filename}.md"
        try:
            with open(description_md_file_path) as file:
                default_context = {"name": self.name, "start_year": self.start_year, "end_year": self.end_year}
                template_context = {**default_context, **context}

                template = file.read()
                template = template.format_map(FormatMapDict(template_context))
                return str(md.render(template))
        except FileNotFoundError:
            return None

    @cached_property
    def description_text_as_html(self) -> Optional[str]:
        """Get the description of the court (where present)."""
        return self.render_markdown_text("description")

    @cached_property
    def historic_documents_support_text_as_html(self) -> Optional[str]:
        """Get support information (where present) on accessing historic court documents not held in FCL."""
        return self.render_markdown_text("historic_docs")

    def __repr__(self) -> str:
        return self.name


class CourtWithJurisdiction(Court):
    def __init__(self, court: Court, jurisdiction: Jurisdiction) -> None:
        self.court: Court = court
        self.jurisdiction: Jurisdiction = jurisdiction
        self.jurisdictions: list[Jurisdiction] = []

    @property
    def code(self) -> CourtCode:
        return CourtCode("/".join((self.court.code, self.jurisdiction.code)))

    @property
    def name(self) -> str:
        return "%s – %s" % (self.court.name, self.jurisdiction.name)

    @property
    def grouped_name(self) -> str:
        return self.court.grouped_name

    @property
    def link(self) -> str:
        return self.court.link

    @property
    def identifier_iri(self) -> str:
        return self.court.identifier_iri

    @property
    def ncn_pattern(self) -> Optional[NeutralCitationPattern]:
        return self.court.ncn_pattern

    @property
    def ncn_examples(self) -> Optional[list[str]]:
        return self.court.ncn_examples

    @property
    def canonical_param(self) -> Optional[CourtParam]:
        return self.court.canonical_param

    @property
    def param_aliases(self) -> list[CourtParam]:
        return self.court.param_aliases

    @property
    def start_year(self) -> Optional[int]:
        return self.court.start_year

    @property
    def end_year(self) -> int:
        return self.court.end_year

    @property
    def jurisdiction_prefix(self) -> Optional[str]:
        return self.jurisdiction.prefix


class CourtGroup:
    def __init__(self, name: Optional[str], courts: list[Court]) -> None:
        self.name: Optional[str] = name
        self.courts: list[Court] = courts

    @property
    def display_heading(self) -> bool:
        """Is this group displayed in the PUI?"""
        return self.name is not None

    def __repr__(self) -> str:
        return f"CourtGroup({self.name!r}, {self.courts!r})"


class CourtNotFoundException(Exception):
    pass


class CourtsRepository:
    def __init__(self, data: RawCourtRepository):
        self._data: RawCourtRepository = data
        self._byParam: dict[CourtParam, Court] = {}
        self._byCode: dict[CourtCode, Court] = {}
        for group in self._data:
            for courtData in group.get("courts", []):
                court = Court(
                    courtData, type=InstitutionType.TRIBUNAL if group["is_tribunal"] else InstitutionType.COURT
                )
                if "param" in courtData:
                    self._byParam[CourtParam(courtData["param"])] = court
                self._byCode[CourtCode(courtData["code"])] = court

    def __repr__(self) -> str:
        return f"CourtsRepository({self._data!r})"

    def get_by_param(self, param: CourtParam) -> Court:
        try:
            return self._byParam[param]
        except KeyError:
            raise CourtNotFoundException()

    def get_court_by_code(self, code: CourtCode) -> Court:
        try:
            return self._byCode[code]
        except KeyError:
            raise CourtNotFoundException()

    def get_court_with_jurisdiction_by_code(
        self, court_code: CourtCode, jursidiction_code: JurisdictionCode
    ) -> CourtWithJurisdiction:
        court = self.get_court_by_code(court_code)
        if court is None:
            raise CourtNotFoundException()
        jurisdiction = court.get_jurisdiction(jursidiction_code)
        if jurisdiction is None:
            raise CourtNotFoundException()
        return CourtWithJurisdiction(court, jurisdiction)

    def get_by_code(self, code: CourtCode) -> Court:
        if "/" in code:
            (court_code, jurisdiction_code) = code.split("/")
            return self.get_court_with_jurisdiction_by_code(CourtCode(court_code), JurisdictionCode(jurisdiction_code))
        else:
            return self.get_court_by_code(code)

    def get_all(self, with_jurisdictions: bool = False) -> list[Court]:
        courts = [
            Court(court, type=InstitutionType.TRIBUNAL if category["is_tribunal"] else InstitutionType.COURT)
            for category in self._data
            for court in category.get("courts", [])
        ]
        if with_jurisdictions:
            return [c for court in courts for c in court.expand_jurisdictions()]
        else:
            return courts

    def get_selectable(self) -> list[Court]:
        courts = []
        for category in self._data:
            for court in category.get("courts", []):
                if court["selectable"]:
                    courts.append(
                        Court(
                            court, type=InstitutionType.TRIBUNAL if category["is_tribunal"] else InstitutionType.COURT
                        )
                    )
        return courts

    def get_selectable_groups(self) -> list[CourtGroup]:
        groups = []
        for category in self._data:
            courts = [
                Court(court, type=InstitutionType.TRIBUNAL if category["is_tribunal"] else InstitutionType.COURT)
                for court in category.get("courts", [])
                if court["selectable"]
            ]
            if len(courts) > 0:
                groups.append(CourtGroup(category.get("display_name"), courts))
        return groups

    def get_grouped_selectable_courts(self) -> list[CourtGroup]:
        groups = []
        for category in self._data:
            if not category.get("is_tribunal"):
                courts = [
                    Court(court, type=InstitutionType.COURT)
                    for court in category.get("courts", [])
                    if court["selectable"]
                ]
                if len(courts) > 0:
                    groups.append(CourtGroup(category.get("display_name"), courts))
        return groups

    def get_grouped_selectable_tribunals(self) -> list[CourtGroup]:
        groups = []
        for category in self._data:
            if category.get("is_tribunal"):
                courts = [
                    Court(court, InstitutionType.TRIBUNAL)
                    for court in category.get("courts", [])
                    if court["selectable"]
                ]
                if len(courts) > 0:
                    groups.append(CourtGroup(category.get("display_name"), courts))
        return groups

    def get_listable_groups(self) -> list[CourtGroup]:
        groups = []
        for category in self._data:
            courts = [
                Court(court, type=InstitutionType.TRIBUNAL if category["is_tribunal"] else InstitutionType.COURT)
                for court in category.get("courts", [])
                if court["listable"]
            ]
            if len(courts) > 0:
                groups.append(CourtGroup(category.get("display_name"), courts))
        return groups

    def get_listable_courts(self) -> list[Court]:
        courts = []
        for group in self._data:
            if not group.get("is_tribunal"):
                for court in group.get("courts", []):
                    if court["listable"]:
                        courts.append(Court(court, InstitutionType.COURT))
        return courts

    def get_listable_tribunals(self) -> list[Court]:
        courts = []
        for group in self._data:
            if group.get("is_tribunal"):
                for court in group.get("courts", []):
                    if court["listable"]:
                        courts.append(Court(court, InstitutionType.TRIBUNAL))
        return courts


yaml = YAML()
datafile = pathlib.Path(__file__).parent / "data/court_names.yaml"
with open(datafile) as f:
    court_data: RawCourtRepository = yaml.load(f)

courts = CourtsRepository(court_data)



import typing

from .courts import Court, InstitutionType
from .types.courts_schema_autogen import RawCourt, RawCourtRepository


class CourtFactory(Court):
    def __init__(self, data, type=InstitutionType.COURT):
        data = make_court_valid(data)
        super().__init__(data, type)


def make_court_valid(data) -> RawCourt:
    for keyword in ["code", "name", "link"]:
        if keyword not in data:
            data[keyword] = f"placeholder {keyword}"
    for keyword in ["selectable", "listable"]:
        if keyword not in data:
            data[keyword] = True

    return typing.cast(RawCourt, data)


def make_court_repo_valid(data) -> RawCourtRepository:
    new_court_groups = []
    for court_group in data:
        new_courts = []
        for court in court_group["courts"]:
            new_courts.append(make_court_valid(court))
        court_group["court"] = new_courts
        new_court_groups.append(court_group)
    return typing.cast(RawCourtRepository, new_court_groups)



# Automatically generated file from a JSON schema


from typing import Required, TypedDict


NeutralCitationNumber = str
r"""
Neutral Citation Number.

An example Neutral Citation Number, matching our expected generic NCN pattern.

pattern: ^\[([0-9]{4})\] ([a-zA-Z]+)(?: ([a-zA-Z]+))? ([0-9]+)(?: \(([a-zA-Z0-9]+)\))?$
"""



class RawCourt(TypedDict, total=False):
    r"""
    Raw Court.

    dependentRequired:
      end_year:
      - start_year
      ncn_examples:
      - ncn_pattern
      ncn_pattern:
      - ncn_examples
    allOf:
      - if:
          properties:
            selectable:
              const: true
        then:
          required:
          - param
      - if:
          properties:
            listable:
              const: true
        then:
          required:
          - param
    """

    code: Required[str]
    r"""
    pattern: ^[A-Za-z]{2,}(-[A-Za-z0-9]+)*$

    Required property
    """

    name: Required[str]
    r""" Required property """

    long_name: str
    grouped_name: str
    param: str
    r""" pattern: ^[a-z]{2,}(?:/[a-z0-9]+)?$ """

    extra_params: list["_RawCourtExtraParamsItem"]
    ncn_pattern: str
    r"""
    Neutral Citation Pattern.

    A regular expression pattern which matches valid NCNs from this court.
    """

    ncn_examples: list["NeutralCitationNumber"]
    r"""
    Neutral Citation examples.

    An array of example NCNs for this court.
    """

    link: Required[str]
    r"""
    format: uri

    Required property
    """

    identifier_iri: str
    r""" format: uri """

    start_year: int
    r""" minimum: 1066 """

    end_year: int
    listable: Required[bool]
    r""" Required property """

    selectable: Required[bool]
    r""" Required property """

    jurisdictions: list["RawJurisdiction"]


class RawCourtGroup(TypedDict, total=False):
    r""" Raw Court Group. """

    name: Required[str]
    r""" Required property """

    display_name: Required[str | None]
    r""" Required property """

    is_tribunal: Required[bool]
    r""" Required property """

    courts: Required[list["RawCourt"]]
    r""" Required property """



RawCourtRepository = list["RawCourtGroup"]
r"""
Raw Court List.

A list of courts
"""



class RawJurisdiction(TypedDict, total=False):
    r""" Raw Jurisdiction. """

    prefix: Required[str]
    r""" Required property """

    name: Required[str]
    r""" Required property """

    code: Required[str]
    r""" Required property """



_RawCourtExtraParamsItem = str
r""" pattern: ^[a-z]{2,}(/[a-z]+)?$ """




from re import compile

from .courts import courts
from .neutral import neutral_url
from .types import NeutralCitationString


def test_court_ncn_example_integrity():
    """Check that each court's example NCNs match that court's NCN pattern."""

    for court in courts.get_all():
        if (
            court.ncn_pattern and court.ncn_examples
        ):  # The schema for court data enforces both of these being present, but check anyway to prevent test issues
            pattern = compile(court.ncn_pattern)
            for ncn_example in court.ncn_examples:
                assert pattern.match(ncn_example)


def test_all_example_ncns_have_conversion_pattern():
    """Check that each example NCN in the court data will convert to a valid URL pattern."""

    for court in courts.get_all():
        if court.ncn_examples:
            for ncn_example in court.ncn_examples:
                assert neutral_url(NeutralCitationString(ncn_example))



import pathlib
import unittest
from datetime import date
from unittest.mock import MagicMock, PropertyMock, mock_open, patch

from ruamel.yaml import YAML

from ds_caselaw_utils.factory import CourtFactory, make_court_repo_valid

from .courts import (
    Court,
    CourtCode,
    CourtGroup,
    CourtNotFoundException,
    CourtParam,
    CourtsRepository,
    CourtWithJurisdiction,
    InstitutionType,
    courts,
)


def mock_with_properties(properties={}):
    mock = MagicMock(spec=dict)
    for property, value in properties.items():
        property_mock = PropertyMock(return_value=value)
        setattr(type(mock), property, property_mock)
        setattr(mock, "mock_%s" % property, property_mock)
    return mock


class TestCourtsRepository(unittest.TestCase):
    def test_loads_all_courts_without_jurisdictions(self):
        data = [
            {
                "name": "court_group",
                "display_name": "court group 1",
                "is_tribunal": False,
                "courts": [{"name": "court1", "jurisdictions": [{"name": "jurisdiction1", "code": "code"}]}],
            }
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        courts = repo.get_all()
        self.assertIn("court1", [c.name for c in courts])
        self.assertNotIn("court1 – jurisdiction1", [c.name for c in courts])

    def test_loads_all_courts_with_jurisdictions(self):
        data = [
            {
                "name": "court_group",
                "display_name": "court group 1",
                "is_tribunal": False,
                "courts": [{"name": "court1", "jurisdictions": [{"name": "jurisdiction1", "code": "code"}]}],
            }
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        courts = repo.get_all(with_jurisdictions=True)
        self.assertIn("court1", [c.name for c in courts])
        self.assertIn("court1 – jurisdiction1", [c.name for c in courts])

    def test_loads_selectable_courts(self):
        data = [
            {
                "name": "court_group",
                "display_name": "court group 1",
                "is_tribunal": False,
                "courts": [
                    {
                        "name": "court1",
                        "selectable": True,
                    },
                    {"name": "court2", "selectable": False},
                ],
            },
            {
                "name": "court_group2",
                "display_name": "court group 2",
                "is_tribunal": False,
                "courts": [{"name": "court3", "selectable": False}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        selectable = repo.get_selectable()
        self.assertIn("court1", [c.name for c in selectable])
        self.assertNotIn("court2", [c.name for c in selectable])
        groups = repo.get_selectable_groups()
        self.assertIn("court group 1", [g.name for g in groups])
        self.assertNotIn("court group 2", [g.name for g in groups])
        self.assertIn("court1", [c.name for g in groups for c in g.courts])
        self.assertNotIn("court2", [c.name for g in groups for c in g.courts])
        self.assertNotIn("court3", [c.name for g in groups for c in g.courts])

    def test_loads_listable_courts(self):
        data = [
            {
                "name": "court_group1",
                "display_name": "court group 1",
                "is_tribunal": False,
                "courts": [
                    {
                        "name": "court1",
                        "listable": True,
                    },
                    {"name": "court2", "listable": False},
                ],
            },
            {
                "name": "court_group2",
                "display_name": "court group 2",
                "is_tribunal": False,
                "courts": [{"name": "court3", "listable": False}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        groups = repo.get_listable_groups()
        self.assertIn("court group 1", [g.name for g in groups])
        self.assertNotIn("court group 2", [g.name for g in groups])
        self.assertIn("court1", [c.name for g in groups for c in g.courts])
        self.assertNotIn("court2", [c.name for g in groups for c in g.courts])
        self.assertNotIn("court3", [c.name for g in groups for c in g.courts])

    def test_loads_court_by_param(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [{"param": "court1", "name": "Court 1"}],
            },
            {
                "name": "court_group2",
                "is_tribunal": False,
                "courts": [{"param": "court2", "name": "Court 2"}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertEqual("Court 2", repo.get_by_param(CourtParam("court2")).name)

    def test_raises_on_unknown_court_param(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [{"param": "court1", "name": "Court 1"}],
            },
            {
                "name": "court_group2",
                "is_tribunal": False,
                "courts": [{"param": "court2", "name": "Court 2"}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertRaises(CourtNotFoundException, repo.get_by_param, "court3")

    def test_loads_court_by_code(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [{"code": "court1", "name": "Court 1"}],
            },
            {
                "name": "court_group2",
                "is_tribunal": False,
                "courts": [{"code": "court2", "name": "Court 2"}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertEqual("Court 2", repo.get_by_code(CourtCode("court2")).name)

    def test_loads_court_with_jurisdiction_by_code(self):
        data = [
            {
                "name": "court_group",
                "is_tribunal": False,
                "courts": [
                    {
                        "code": "court1",
                        "name": "Court 1",
                        "jurisdictions": [{"code": "jurisdiction1", "name": "Jurisdiction 1"}],
                    }
                ],
            }
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertEqual("Court 1 – Jurisdiction 1", repo.get_by_code(CourtCode("court1/jurisdiction1")).name)

    def test_raises_error_for_nonexistent_jurisdictions(self):
        data = [
            {
                "name": "court_group",
                "is_tribunal": False,
                "courts": [
                    {
                        "code": "court1",
                        "name": "Court 1",
                        "jurisdictions": [{"code": "jurisdiction1", "name": "Jurisdiction 1"}],
                    }
                ],
            }
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertRaises(CourtNotFoundException, repo.get_by_code, "court1/jurisdiction2")
        self.assertRaises(CourtNotFoundException, repo.get_by_code, "court2/jurisdiction1")

    def test_raises_on_unknown_court_code(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [{"code": "court1", "name": "Court 1"}],
            },
            {
                "name": "court_group2",
                "is_tribunal": False,
                "courts": [{"code": "court2", "name": "Court 2"}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertRaises(CourtNotFoundException, repo.get_by_code, "court3")

    def test_returns_listable_courts(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [
                    {"param": "court1", "listable": True, "name": "Court 1"},
                    {"param": "court2", "listable": False, "name": "Court 2"},
                ],
            },
            {
                "name": "court_group2",
                "is_tribunal": True,
                "courts": [{"param": "court3", "listable": True, "name": "Court 3"}],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertIn("court1", [c.canonical_param for c in repo.get_listable_courts()])
        self.assertNotIn("court2", [c.canonical_param for c in repo.get_listable_courts()])
        self.assertNotIn("court3", [c.canonical_param for c in repo.get_listable_courts()])

    def test_returns_listable_tribunals(self):
        data = [
            {
                "name": "court_group1",
                "is_tribunal": False,
                "courts": [
                    {"param": "court1", "listable": True, "name": "Court 1"},
                ],
            },
            {
                "name": "court_group2",
                "is_tribunal": True,
                "courts": [
                    {"param": "court2", "listable": False, "name": "Court 2"},
                    {"param": "court3", "listable": True, "name": "Court 3"},
                ],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        self.assertNotIn("court1", [c.canonical_param for c in repo.get_listable_tribunals()])
        self.assertNotIn("court2", [c.canonical_param for c in repo.get_listable_tribunals()])
        self.assertIn("court3", [c.canonical_param for c in repo.get_listable_tribunals()])

    def test_returns_grouped_selectable_courts(self):
        data = [
            {
                "name": "group2",
                "display_name": "Court group",
                "is_tribunal": False,
                "courts": [
                    {"param": "court1", "selectable": True, "name": "Selectable court"},
                    {
                        "param": "court2",
                        "selectable": False,
                        "name": "Unselectable court",
                    },
                ],
            },
            {
                "name": "group2",
                "display_name": "Tribunal group",
                "is_tribunal": True,
                "courts": [
                    {
                        "param": "court3",
                        "selectable": True,
                        "name": "Selectable tribunal",
                    }
                ],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        groups = repo.get_grouped_selectable_courts()
        self.assertIn("Court group", [g.name for g in groups])
        self.assertNotIn("Tribunal group", [g.name for g in groups])
        self.assertIn("Selectable court", [c.name for g in groups for c in g.courts])
        self.assertNotIn("Unselectable court", [c.name for g in groups for c in g.courts])
        self.assertNotIn("Selectable tribunal", [c.name for g in groups for c in g.courts])

    def test_returns_grouped_selectable_tribunals(self):
        data = [
            {
                "name": "group1",
                "display_name": "Court group",
                "is_tribunal": False,
                "courts": [
                    {"param": "court1", "selectable": True, "name": "Selectable court"},
                ],
            },
            {
                "name": "group2",
                "display_name": "Tribunal group",
                "is_tribunal": True,
                "courts": [
                    {
                        "param": "court2",
                        "selectable": True,
                        "name": "Selectable tribunal",
                    },
                    {
                        "param": "court3",
                        "selectable": False,
                        "name": "Unselectable tribunal",
                    },
                ],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        groups = repo.get_grouped_selectable_tribunals()
        self.assertIn("Tribunal group", [g.name for g in groups])
        self.assertNotIn("Court group", [g.name for g in groups])
        self.assertIn("Selectable tribunal", [c.name for g in groups for c in g.courts])
        self.assertNotIn("Unselectable tribunal", [c.name for g in groups for c in g.courts])
        self.assertNotIn("Selectable court", [c.name for g in groups for c in g.courts])

    def test_repr(self):
        data = [
            {
                "name": "group1",
                "display_name": "Court group",
                "is_tribunal": False,
                "courts": [
                    {"param": "court1", "selectable": True, "name": "Selectable court"},
                ],
            }
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        assert "'name': 'group1', 'display_name': 'Court group'" in str(repo)

    def test_identifies_institution_type(self):
        data = [
            {
                "name": "courts",
                "is_tribunal": False,
                "courts": [
                    {"param": "court", "listable": True, "name": "Court"},
                ],
            },
            {
                "name": "tribunals",
                "is_tribunal": True,
                "courts": [
                    {"param": "tribunal", "listable": True, "name": "Tribunal"},
                ],
            },
        ]
        valid_data = make_court_repo_valid(data)
        repo = CourtsRepository(valid_data)
        institutions = repo.get_all()

        assert institutions[0].type is InstitutionType.COURT
        assert institutions[1].type is InstitutionType.TRIBUNAL


class TestCourt(unittest.TestCase):
    def test_repr_string(self):
        court = CourtFactory({"name": "court_name"})
        self.assertEqual("court_name", str(court))
        self.assertEqual("court_name", repr(court))

    def test_grouped_name_explicit(self):
        court = CourtFactory({"grouped_name": "court_name"})
        self.assertEqual("court_name", court.grouped_name)

    def test_grouped_name_default(self):
        court = CourtFactory({"name": "court_name"})
        self.assertEqual("court_name", court.grouped_name)

    def test_param_aliases(self):
        court = CourtFactory({"param": "param_1", "extra_params": ["param_2"]})
        self.assertEqual(["param_1", "param_2"], court.param_aliases)

    def test_end_year_explicit(self):
        court = CourtFactory({"end_year": 1983})
        self.assertEqual(1983, court.end_year)

    def test_end_year_default(self):
        court = CourtFactory({})
        self.assertEqual(date.today().year, court.end_year)

    def test_type_when_court(self):
        court = CourtFactory({})
        assert court.type is InstitutionType.COURT

    def test_type_when_tribunal(self):
        court = CourtFactory({}, type=InstitutionType.TRIBUNAL)
        assert court.type is InstitutionType.TRIBUNAL

    def test_get_jurisdiction(self):
        court = CourtFactory({"jurisdictions": [{"code": "jurisdiction1", "name": "Jurisdiction 1"}]})
        jurisdiction = court.get_jurisdiction("jurisdiction1")
        assert jurisdiction
        self.assertEqual("Jurisdiction 1", jurisdiction.name)

    def test_get_nonexistent_jurisdiction(self):
        court = CourtFactory({"jurisdictions": [{"code": "jurisdiction1", "name": "Jurisdiction 1"}]})
        jurisdiction = court.get_jurisdiction("jurisdiction2")
        self.assertIsNone(jurisdiction)

    def test_expand_jurisdictions(self):
        court = CourtFactory(
            {
                "name": "Court 1",
                "jurisdictions": [{"code": "jurisdiction1", "name": "Jurisdiction 1"}],
            }
        )
        expanded = court.expand_jurisdictions()
        self.assertIn("Court 1", [c.name for c in expanded])
        self.assertIn("Court 1 – Jurisdiction 1", [c.name for c in expanded])
        for c in expanded:
            assert issubclass(type(c), Court)

    def test_render_markdown_text_if_no_canonical_param(self):
        court = CourtFactory({"param": "test"})
        assert court.description_text_as_html is None

    def test_render_markdown_text_if_no_file(self):
        court = CourtFactory({"param": "test"})
        assert court.description_text_as_html is None

    def test_render_markdown_text_if_file(self):
        court = CourtFactory({"param": "test"})
        with patch("pathlib.Path.is_file", True), patch("builtins.open", mock_open(read_data="**Test** description.")):
            assert court.render_markdown_text("test") == "<p><strong>Test</strong> description.</p>\n"

    def test_render_markdown_text_with_context(self):
        court = CourtFactory({"param": "test", "name": "test name", "start_year": 2000, "end_year": 2025})
        with (
            patch("pathlib.Path.is_file", True),
            patch(
                "builtins.open",
                mock_open(
                    read_data="**Test** description.\n - Name: {name}\n - Start year: {start_year}\n - End year: {end_year}\n - Do not replace: {do_not_replace}"
                ),
            ),
        ):
            assert (
                court.render_markdown_text("test")
                == "<p><strong>Test</strong> description.</p>\n<ul>\n<li>Name: test name</li>\n<li>Start year: 2000</li>\n<li>End year: 2025</li>\n<li>Do not replace: {do_not_replace}</li>\n</ul>\n"
            )

    @patch("ds_caselaw_utils.courts.Court.render_markdown_text")
    def test_description_text_as_html(self, mock_render):
        court = CourtFactory({"param": "test"})
        mock_render.return_value = "<p>Description.</p>"

        assert court.description_text_as_html == "<p>Description.</p>"
        mock_render.assert_called_once_with("description")

    @patch("ds_caselaw_utils.courts.Court.render_markdown_text")
    def test_historic_documents_support_text_as_html(self, mock_render):
        court = CourtFactory({"param": "test"})
        mock_render.return_value = "<p>Support text.</p>"

        assert court.historic_documents_support_text_as_html == "<p>Support text.</p>"
        mock_render.assert_called_once_with("historic_docs")


class TestCourtWithJurisdiction(unittest.TestCase):
    def test_code(self):
        # It appends the jurisdiction code to the court code
        # separated with a slash
        court = mock_with_properties({"code": "court_code"})
        jurisdiction = mock_with_properties({"code": "jurisdiction_code"})
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.code, "court_code/jurisdiction_code")
        court.mock_code.assert_called()
        jurisdiction.mock_code.assert_called()

    def test_name(self):
        # It appends the jurisdiction name to the court name
        # separated with an en-dash
        court = mock_with_properties({"name": "court_name"})
        jurisdiction = mock_with_properties({"name": "jurisdiction_name"})
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.name, "court_name – jurisdiction_name")
        court.mock_name.assert_called()
        jurisdiction.mock_name.assert_called()

    def test_grouped_name(self):
        # It returns the court grouped name
        court = mock_with_properties({"grouped_name": "court_grouped_name"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.grouped_name, "court_grouped_name")
        court.mock_grouped_name.assert_called()

    def test_link_and_iri_with_default_iri(self):
        # It returns the court link
        court = CourtFactory({"link": "court_link"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.link, "court_link")
        self.assertEqual(cwj.identifier_iri, "court_link")

    def test_link_and_iri_with_custom_iri(self):
        # It returns the court link
        court = CourtFactory({"link": "court_link", "identifier_iri": "court_identifier_iri"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.link, "court_link")
        self.assertEqual(cwj.identifier_iri, "court_identifier_iri")

    def test_ncn_pattern(self):
        # It returns the court's NCN pattern
        court = mock_with_properties({"ncn_pattern": "court_ncn"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.ncn_pattern, "court_ncn")
        court.mock_ncn_pattern.assert_called()

    def test_ncn_examples(self):
        # It returns the court's NCN examples
        court = mock_with_properties({"ncn_examples": ["[2025] UKSC 123"]})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.ncn_examples, ["[2025] UKSC 123"])
        court.mock_ncn_examples.assert_called()

    def test_canonical_param(self):
        # It returns the court canonical_param
        court = mock_with_properties({"canonical_param": "court_canonical_param"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.canonical_param, "court_canonical_param")
        court.mock_canonical_param.assert_called()

    def param_aliases(self):
        # It returns the court param_aliases
        court = mock_with_properties({"param_aliases": "court_param_aliases"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.param_aliases, "court_param_aliases")
        court.mock_param_aliases.assert_called()

    def start_year(self):
        # It returns the court start_year
        court = mock_with_properties({"start_year": "court_start_year"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.start_year, "court_start_year")
        court.mock_start_year.assert_called()

    def end_year(self):
        # It returns the court end_year
        court = mock_with_properties({"end_year": "court_end_year"})
        jurisdiction = mock_with_properties()
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.end_year, "court_end_year")
        court.mock_end_year.assert_called()

    def test_jurisdiction_prefix(self):
        # It returns the jurisdiction prefix:
        court = mock_with_properties()
        jurisdiction = mock_with_properties({"prefix": "ABC"})
        cwj = CourtWithJurisdiction(court, jurisdiction)
        self.assertEqual(cwj.jurisdiction_prefix, "ABC")
        jurisdiction.mock_prefix.assert_called()


class TestCourtGroup(unittest.TestCase):
    def test_display_heading_when_has_display_name(self):
        group = CourtGroup("name", [])
        assert group.display_heading

    def test_dont_display_heading_when_no_display_name(self):
        group = CourtGroup(None, [])
        assert not group.display_heading

    def test_repr(self):
        group = CourtGroup("name", [])
        assert str(group) == "CourtGroup('name', [])"


class TestCourts(unittest.TestCase):
    def test_loads_court_yaml(self):
        yaml = YAML()
        datafile = pathlib.Path(__file__).parent / "data/court_names.yaml"
        with open(datafile) as f:
            court_data = yaml.load(f)
        courts_from_yaml = []
        for group in court_data:
            for court in group.get("courts"):
                courts_from_yaml.append(court)
        for court, data in zip(courts.get_all(), courts_from_yaml):
            self.assertEqual(court.name, data["name"])


class TestAutogenCourtsJson(unittest.TestCase):
    def test_autogen_courts_json(self):
        """Check that one edge case is handled correctly in the autogenerated JSON"""
        datafile = datafile = pathlib.Path(__file__).parent.parent.parent / "src/autogen/courts.json"
        with open(datafile) as f:
            raw_json = f.read()
        assert '"identifier_iri": "http://www.tribunals.gov.uk/ImmigrationAsylum/"' in raw_json



from re import Pattern
from typing import NewType

# Types which are used for identifying courts and their jurisdictions
CourtCode = NewType("CourtCode", str)
CourtParam = NewType("CourtParam", str)
JurisdictionCode = NewType("JurisdictionCode", str)

# Types which are used for identifying a particular detail of a court
NeutralCitationPattern = NewType("NeutralCitationPattern", Pattern[str])

# Types which are used to identify a judgment
NeutralCitationString = NewType("NeutralCitationString", str)
NCNBasedUriString = NewType("NCNBasedUriString", str)



import pytest

from .neutral import neutral_url
from .types import NeutralCitationString


class TestNeutralURL:
    @pytest.mark.parametrize(
        "ncn, expected_uri",
        [
            ("[2022] UKSC 1", "uksc/2022/1"),
            ("[1604] EWCA Crim 555", "ewca/crim/1604/555"),
            ("[2022] EWHC 1 (Comm)", "ewhc/comm/2022/1"),
            ("[1999] EWCOP 7", "ewcop/1999/7"),
            ("[2022] UKUT 1 (IAC)", "ukut/iac/2022/1"),
            ("[2022] EAT 1", "eat/2022/1"),
            ("[2022] UKFTT 1 (TC)", "ukftt/tc/2022/1"),
            ("[2022] UKFTT 1 (GRC)", "ukftt/grc/2022/1"),
            ("[2025] UKFTT 1 (PC)", "ukftt/pc/2025/1"),
            ("[2022] EWHC 1 (KB)", "ewhc/kb/2022/1"),
            ("[2023] UKAIT 1", "ukait/2023/1"),
            ("[2024] EWCOP 17 (T2)", "ewcop/t2/2024/17"),
            ("[2000] UKIPTrib 99", "ukiptrib/2000/99"),
            ("[2000] EWCR 99", "ewcr/2000/99"),
            ("[1999] UKFTT 1 (HESC)", "ukftt/hesc/1999/1"),
        ],
    )
    def test_good_neutral_urls(self, ncn, expected_uri):
        assert neutral_url(NeutralCitationString(ncn)) == expected_uri

    @pytest.mark.parametrize(
        "ncn",
        [
            "",
            "1604] EWCA Crim 555",
            "[2022 EWHC 1 Comm",
            "[1999] EWCOP",
            "[2022] UKUT B1 IAC",  # Could be a Bailii reference, might want to drop B in future.
            "[2022] EAT A",
            "[2022] NOTACOURT 1 TC",
            "[2022] EWHC 1 (T2)",
            "[2000] EWCRC 99",
        ],
    )
    def test_bad_neutral_urls(self, ncn):
        assert neutral_url(NeutralCitationString(ncn)) is None



from .factory import make_court_repo_valid


def test_factory():
    input_data = [
        {
            "name": "court_group",
            "display_name": "court group 1",
            "courts": [
                {
                    "name": "court1",
                    "selectable": True,
                },
                {"name": "court2", "selectable": False},
            ],
        },
        {
            "name": "court_group2",
            "display_name": "court group 2",
            "courts": [{"name": "court3", "selectable": False}],
        },
    ]

    output_data = [
        {
            "name": "court_group",
            "display_name": "court group 1",
            "courts": [
                {
                    "name": "court1",
                    "selectable": True,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                },
                {
                    "name": "court2",
                    "selectable": False,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                },
            ],
            "court": [
                {
                    "name": "court1",
                    "selectable": True,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                },
                {
                    "name": "court2",
                    "selectable": False,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                },
            ],
        },
        {
            "name": "court_group2",
            "display_name": "court group 2",
            "courts": [
                {
                    "name": "court3",
                    "selectable": False,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                }
            ],
            "court": [
                {
                    "name": "court3",
                    "selectable": False,
                    "code": "placeholder code",
                    "link": "placeholder link",
                    "listable": True,
                }
            ],
        },
    ]
    assert str(output_data) == str(make_court_repo_valid(input_data))
